@@ -58,7 +58,55 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
 
     5. Use the low-level API (:py:func:`get_p_template` and :py:func:`set_p_fun`) or high level API (:py:func:`set_uncertainty_values`) to create scenarios for robust MPC (optional).
 
-    6. Finally, call :py:func:`setup`.
+    6. Use :py:meth:`get_tvp_template` and :py:meth:`set_tvp_fun` to create a method to obtain new time-varying parameters at each iteration.
+
+    7. To finalize the class configuration there are two routes. The default approach is to call :py:meth:`setup`. For deep customization use the combination of :py:meth:`prepare_nlp` and :py:meth:`create_nlp`. See graph below for an illustration of the process.
+
+    .. graphviz::
+        :name: route_to_setup
+        :caption: Route to setting up the MPC class.
+        :align: center
+
+        digraph G {
+            graph [fontname = "helvetica"];
+            rankdir=LR;
+
+            subgraph cluster_main {
+                node [fontname = "helvetica", shape=box, fontcolor="#404040", color="#707070"];
+                edge [fontname = "helvetica", color="#707070"];
+
+                start [label="Two ways to setup"];
+                setup [label="setup", href="../api/do_mpc.controller.MPC.setup.html", target="_top", fontname = "Consolas"];
+                create_nlp [label="create_nlp", href="../api/do_mpc.controller.MPC.create_nlp.html", target="_top", fontname = "Consolas"];
+                process [label="Modify NLP"];
+                prepare_nlp [label="prepare_nlp", href="../api/do_mpc.controller.MPC.prepare_nlp.html", target="_top", fontname = "Consolas"];
+                finish [label="Configured MPC class"]
+                start -> setup, prepare_nlp;
+                prepare_nlp -> process;
+                process -> create_nlp;
+                setup, create_nlp -> finish;
+                color=none;
+            }
+
+            subgraph cluster_modification {
+                rankdir=TB;
+                node [fontname = "helvetica", shape=box, fontcolor="#404040", color="#707070"];
+                edge [fontname = "helvetica", color="#707070"];
+                opt_x [label="opt_x", href="../api/do_mpc.controller.MPC.opt_x.html", target="_top", fontname = "Consolas"];
+                opt_p [label="opt_p", href="../api/do_mpc.controller.MPC.opt_p.html", target="_top", fontname = "Consolas"];
+                nlp_cons [label="nlp_cons", href="../api/do_mpc.controller.MPC.nlp_cons.html", target="_top", fontname = "Consolas"];
+                nlp_obj [label="nlp_obj", href="../api/do_mpc.controller.MPC.nlp_obj.html", target="_top", fontname = "Consolas"];
+
+                opt_x -> nlp_cons, nlp_obj;
+                opt_p -> nlp_cons, nlp_obj;
+
+                label = "Attributes to modify the NLP.";
+		        color=black;
+            }
+
+            nlp_cons -> process;
+            nlp_obj -> process;
+        }
 
     .. warning::
 
@@ -107,6 +155,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
             'collocation_deg',
             'collocation_ni',
             'nl_cons_check_colloc_points',
+            'nl_cons_single_slack',
             'cons_check_colloc_points',
             'store_full_solution',
             'store_lagr_multiplier',
@@ -123,6 +172,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         self.collocation_deg = 2
         self.collocation_ni = 1
         self.nl_cons_check_colloc_points = False
+        self.nl_cons_single_slack = False
         self.cons_check_colloc_points = True
         self.store_full_solution = False
         self.store_lagr_multiplier = True
@@ -134,14 +184,13 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         self.nlpsol_opts = {} # Will update default options with this dict.
 
         # Flags are checked when calling .setup.
-        self.flags = {
-            'setup': False,
+        self.flags.update({
             'set_objective': False,
             'set_rterm': False,
             'set_tvp_fun': False,
             'set_p_fun': False,
             'set_initial_guess': False,
-        }
+        })
 
     @property
     def opt_x_num(self):
@@ -170,7 +219,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
 
         The attribute can be used **to manually set a custom initial guess or for debugging purposes**.
 
-        ** How to query ``opt_x_num``?**
+        **How to query?**
 
         Querying the structure is more complicated than it seems at first look because of the scenario-tree used
         for robust MPC. To obtain all collocation points for the finite element at time-step :math:`k` and scenario :math:`b` use:
@@ -256,6 +305,97 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
     def opt_p_num(self, val):
         self._opt_p_num = val
 
+    @property
+    def opt_x(self):
+        """Full structure of (symbolic) MPC optimization variables.
+
+        The attribute is a CasADi symbolic structure with nested power indices.
+        It can be indexed as follows:
+
+        ::
+
+            # dynamic states:
+            opt_x['_x', time_step, scenario, collocation_point, _x_name]
+            # algebraic states:
+            opt_x['_z', time_step, scenario, collocation_point, _z_name]
+            # inputs:
+            opt_x['_u', time_step, scenario, _u_name]
+            # slack variables for soft constraints:
+            opt_x['_eps', time_step, scenario, _nl_cons_name]
+
+        The names refer to those given in the :py:class:`do_mpc.model.Model` configuration.
+        Further indices are possible, if the variables are itself vectors or matrices.
+
+        The attribute can be used to alter the objective function or constraints of the NLP.
+
+        **How to query?**
+
+        Querying the structure is more complicated than it seems at first look because of the scenario-tree used
+        for robust MPC. To obtain all collocation points for the finite element at time-step :math:`k` and scenario :math:`b` use:
+
+        ::
+
+            horzcat(*[mpc.opt_x['_x',k,b,-1]]+mpc.opt_x['_x',k+1,b,:-1])
+
+        Due to the multi-stage formulation at any given time :math:`k` we can have multiple future scenarios.
+        However, there is only exactly one scenario that lead to the current node in the tree.
+        Thus the collocation points associated to the finite element :math:`k` lie in the past.
+
+        The concept is illustrated in the figure below:
+
+        .. figure:: ../static/collocation_points_scenarios.svg
+
+        .. note::
+
+            The attribute ``opt_x`` carries the scaled values of all variables.
+
+        .. note::
+
+            The attribute is populated when calling :py:func:`setup` or :py:func:`prepare_nlp`
+        """
+        return self._opt_x
+
+    @opt_x.setter
+    def opt_x(self, val):
+        self._opt_x = val
+
+    @property
+    def opt_p(self):
+        """Full structure of (symbolic) MPC parameters.
+
+        The attribute is a CasADi numeric structure with nested power indices.
+        It can be indexed as follows:
+
+        ::
+
+            # initial state:
+            opt_p['_x0', _x_name]
+            # uncertain scenario parameters
+            opt_p['_p', scenario, _p_name]
+            # time-varying parameters:
+            opt_p['_tvp', time_step, _tvp_name]
+            # input at time k-1:
+            opt_p['_u_prev', time_step, scenario]
+
+        The names refer to those given in the :py:class:`do_mpc.model.Model` configuration.
+        Further indices are possible, if the variables are itself vectors or matrices.
+
+        .. warning::
+
+            Do not tweak or overwrite this attribute unless you known what you are doing.
+
+        .. note::
+
+            The attribute is populated when calling :py:func:`setup` or :py:func:`prepare_nlp`
+
+        """
+        return self._opt_p
+
+    @opt_p.setter
+    def opt_p(self, val):
+        self._opt_p = val
+
+
     @IndexedProperty
     def terminal_bounds(self, ind):
         """Query and set the terminal bounds for the states.
@@ -380,7 +520,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         :param use_terminal_bounds: Choose if terminal bounds for the states are used. Defaults to ``True``. Set terminal bounds with :py:attr:`terminal_bounds`.
         :type use_terminal_bounds: bool
 
-        :param state_discretization: Choose the state discretization for continuous models. Currently only ``'collocation'`` is available. Defaults to ``'collocation'``.
+        :param state_discretization: Choose the state discretization for continuous models. Currently only ``'collocation'`` is available. Defaults to ``'collocation'``. Has no effect if model is created in ``discrete`` type.
         :type state_discretization: str
 
         :param collocation_type: Choose the collocation type for continuous models with collocation as state discretization. Currently only ``'radau'`` is available. Defaults to ``'radau'``.
@@ -395,6 +535,9 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         :param nl_cons_check_colloc_points: For orthogonal collocation choose whether the nonlinear bounds set with :py:func:`set_nl_cons` are evaluated once per finite Element or for each collocation point. Defaults to ``False`` (once per collocation point).
         :type nl_cons_check_colloc_points: bool
 
+        :param nl_cons_single_slack: If ``True``, soft-constraints set with :py:func:`set_nl_cons` introduce only a single slack variable for the entire horizon. Defaults to ``False``.
+        :type nl_cons_single_slack: bool
+
         :param cons_check_colloc_points: For orthogonal collocation choose whether the linear bounds set with :py:attr:`bounds` are evaluated once per finite Element or for each collocation point. Defaults to ``True`` (for all collocation points).
         :type cons_check_colloc_points: bool
 
@@ -415,12 +558,12 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
             ::
 
                 MPC.set_param(nlpsol_opts = {'ipopt.linear_solver': 'MA27'})
-        .. note:: To surpress the output of IPOPT, please use:
+        .. note:: To suppress the output of IPOPT, please use:
 
             ::
 
-                surpress_ipopt = {'ipopt.print_level':0, 'ipopt.sb': 'yes', 'print_time':0}
-                MPC.set_param(nlpsol_opts = surpress_ipopt)
+                suppress_ipopt = {'ipopt.print_level':0, 'ipopt.sb': 'yes', 'print_time':0}
+                MPC.set_param(nlpsol_opts = suppress_ipopt)
         """
         assert self.flags['setup'] == False, 'Setting parameters after setup is prohibited.'
 
@@ -461,34 +604,38 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         """
         assert mterm.shape == (1,1), 'mterm must have shape=(1,1). You have {}'.format(mterm.shape)
         assert lterm.shape == (1,1), 'lterm must have shape=(1,1). You have {}'.format(lterm.shape)
-        assert self.flags['setup'] == False, 'Cannot call .set_objective after .setup_model.'
+        assert self.flags['setup'] == False, 'Cannot call .set_objective after .setup().'
 
         _x, _u, _z, _tvp, _p = self.model['x','u','z','tvp','p']
 
 
         # Check if mterm is valid:
-        if isinstance(mterm, casadi.DM):
-            pass
-        elif isinstance(mterm, (casadi.SX, casadi.MX)):
-            assert set(symvar(mterm)).issubset(set(symvar(vertcat(_x, _p)))), 'mterm must be solely a function of _x and _p.'
-        else:
+        if not isinstance(mterm, (casadi.DM, casadi.SX, casadi.MX)):
             raise Exception('mterm must be of type casadi.DM, casadi.SX or casadi.MX. You have: {}.'.format(type(mterm)))
 
         # Check if lterm is valid:
-        if isinstance(lterm, casadi.DM):
-            pass
-        elif isinstance(lterm, (casadi.SX, casadi.MX)):
-            assert set(symvar(lterm)).issubset(set(symvar(vertcat(_x, _u, _z, _tvp, _p)))), 'lterm must be solely a function of _x, _u, _z, _tvp, _p.'
-        else:
+        if not isinstance(lterm, (casadi.DM, casadi.SX, casadi.MX)):
             raise Exception('lterm must be of type casadi.DM, casadi.SX or casadi.MX. You have: {}.'.format(type(lterm)))
 
         self.mterm = mterm
         # TODO: This function should be evaluated with scaled variables.
-        self.mterm_fun = Function('mterm', [_x, _p], [mterm])
+        self.mterm_fun = Function('mterm', [_x, _tvp, _p], [mterm])
 
         self.lterm = lterm
         self.lterm_fun = Function('lterm', [_x, _u, _z, _tvp, _p], [lterm])
 
+        # Check if lterm and mterm use invalid variables as inputs.
+        # For the check we evaluate the function with dummy inputs and expect a DM output.
+        err_msg = '{} contains invalid symbolic variables as inputs. Must contain only: {}'
+        try:
+            self.mterm_fun(_x(0),_tvp(0),_p(0))
+        except:
+            raise Exception(err_msg.format('mterm','_x, _tvp, _p'))
+        try:
+            self.lterm(_x(0),_u(0), _z(0), _tvp(0), _p(0))
+        except:
+            err_msg.format('lterm', '_x, _u, _z, _tvp, _p')
+
         self.flags['set_objective'] = True
 
     def set_rterm(self, **kwargs):
@@ -533,7 +680,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
             For :math:`k=0` we obtain :math:`u_{-1}` from the previous solution.
 
         """
-        assert self.flags['setup'] == False, 'Cannot call .set_rterm after .setup_model.'
+        assert self.flags['setup'] == False, 'Cannot call .set_rterm after .setup().'
 
         self.flags['set_rterm'] = True
         for key, val in kwargs.items():
@@ -541,7 +688,6 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
             assert isinstance(val, (int, float, np.ndarray)), 'Value for {} must be int, float or numpy.ndarray. You have: {}'.format(key, type(val))
             self.rterm_factor[key] = val
 
-
     def get_p_template(self, n_combinations):
         """Obtain output template for :py:func:`set_p_fun`.
 
@@ -598,7 +744,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         :rtype: None
         """
         self.n_combinations = n_combinations
-        p_template = struct_symSX([
+        p_template = self.model.sv.sym_struct([
             entry('_p', repeat=n_combinations, struct=self.model._p)
         ])
         return p_template(0)
@@ -812,18 +958,10 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
 
         """
 
-        self._setup_nl_cons()
-        self._check_validity()
-        self._setup_mpc_optim_problem()
+        self.prepare_nlp()
+        self.create_nlp()
 
-        # Gather meta information:
-        meta_data = {key: getattr(self, key) for key in self.data_fields}
-        meta_data.update({'structure_scenario': self.scenario_tree['structure_scenario']})
-        self.data.set_meta(**meta_data)
 
-        self._prepare_data()
-
-        self.flags['setup'] = True
 
     def set_initial_guess(self):
         """Initial guess for optimization variables.
@@ -868,7 +1006,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         elif isinstance(x0, structure3.DMStruct):
             x0 = x0.cat
         else:
-            raise Exception('Invalid type {} for x0. Must be {}'.format(type(x0), (np.ndarray, casadi.DM, structre3.DMStruct)))
+            raise Exception('Invalid type {} for x0. Must be {}'.format(type(x0), (np.ndarray, casadi.DM, structure3.DMStruct)))
 
         # Check input shape.
         n_val = np.prod(x0.shape)
@@ -932,27 +1070,44 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         return u0.full()
 
 
-    def _setup_mpc_optim_problem(self):
-        """Private method of the MPC class to construct the MPC optimization problem.
-        The method depends on inherited methods from the :py:class:`do_mpc.optimizer.Optimizer`.
-
-        The :py:class:`do_mpc.estimator.MHE` has a similar method with similar structure.
+    def _prepare_nlp(self):
+        """Internal method. See detailed documentation with optimizer.prepare_nlp
         """
+        nl_cons_input = self.model['x', 'u', 'z', 'tvp', 'p']
+        self._setup_nl_cons(nl_cons_input)
+        self._check_validity()
+
         # Obtain an integrator (collocation, discrete-time) and the amount of intermediate (collocation) points
         ifcn, n_total_coll_points = self._setup_discretization()
         n_branches, n_scenarios, child_scenario, parent_scenario, branch_offset = self._setup_scenario_tree()
+
+        # How many scenarios arise from the scenario tree (robust multi-stage MPC)
         n_max_scenarios = self.n_combinations ** self.n_robust
+
+        # If open_loop option is active, all scenarios (at a given stage) have the same input.
+        if self.open_loop:
+            n_u_scenarios = 1
+        else:
+            # Else: Each scenario has its own input.
+            n_u_scenarios = n_max_scenarios
+
+        # How many slack variables (for soft constraints) are introduced over the horizon.
+        if self.nl_cons_single_slack:
+            n_eps = 1
+        else:
+            n_eps = self.n_horizon
+
         # Create struct for optimization variables:
-        self.opt_x = opt_x = struct_symSX([
+        self._opt_x = opt_x = self.model.sv.sym_struct([
             # One additional point (in the collocation dimension) for the final point.
             entry('_x', repeat=[self.n_horizon+1, n_max_scenarios,
                                 1+n_total_coll_points], struct=self.model._x),
             entry('_z', repeat=[self.n_horizon, n_max_scenarios,
                                 max(n_total_coll_points,1)], struct=self.model._z),
-            entry('_u', repeat=[self.n_horizon, n_max_scenarios], struct=self.model._u),
-            entry('_eps', repeat=[self.n_horizon, n_max_scenarios], struct=self._eps),
+            entry('_u', repeat=[self.n_horizon, n_u_scenarios], struct=self.model._u),
+            entry('_eps', repeat=[n_eps, n_max_scenarios], struct=self._eps),
         ])
-        self.n_opt_x = self.opt_x.shape[0]
+        self.n_opt_x = self._opt_x.shape[0]
         # NOTE: The entry _x[k,child_scenario[k,s,b],:] starts with the collocation points from s to b at time k
         #       and the last point contains the child node
         # NOTE: Currently there exist dummy collocation points for the initial state (for each branch)
@@ -967,9 +1122,9 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
 
 
         # Create struct for optimization parameters:
-        self.opt_p = opt_p = struct_symSX([
+        self._opt_p = opt_p = self.model.sv.sym_struct([
             entry('_x0', struct=self.model._x),
-            entry('_tvp', repeat=self.n_horizon, struct=self.model._tvp),
+            entry('_tvp', repeat=self.n_horizon+1, struct=self.model._tvp),
             entry('_p', repeat=self.n_combinations, struct=self.model._p),
             entry('_u_prev', struct=self.model._u),
         ])
@@ -978,13 +1133,13 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         self.n_opt_p = opt_p.shape[0]
 
         # Dummy struct with symbolic variables
-        self.aux_struct = struct_symSX([
+        self.aux_struct = self.model.sv.sym_struct([
             entry('_aux', repeat=[self.n_horizon, n_max_scenarios], struct=self.model._aux_expression)
         ])
         # Create mutable symbolic expression from the struct defined above.
-        self.opt_aux = opt_aux = struct_SX(self.aux_struct)
+        self._opt_aux = opt_aux = self.model.sv.struct(self.aux_struct)
 
-        self.n_opt_aux = self.opt_aux.shape[0]
+        self.n_opt_aux = opt_aux.shape[0]
 
         self.lb_opt_x = opt_x(-np.inf)
         self.ub_opt_x = opt_x(np.inf)
@@ -1011,6 +1166,9 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
             # For all scenarios (grows exponentially with n_robust)
             for s in range(n_scenarios[k]):
                 # For all childen nodes of each node at stage k, discretize the model equations
+
+                # Scenario index for u is always 0 if self.open_loop = True
+                s_u = 0 if self.open_loop else s
                 for b in range(n_branches[k]):
                     # Obtain the index of the parameter values that should be used for this scenario
                     current_scenario = b + branch_offset[k][s]
@@ -1019,7 +1177,7 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
                     col_xk = vertcat(*opt_x['_x', k+1, child_scenario[k][s][b], :-1])
                     col_zk = vertcat(*opt_x['_z', k, child_scenario[k][s][b]])
                     [g_ksb, xf_ksb] = ifcn(opt_x['_x', k, s, -1], col_xk,
-                                           opt_x['_u', k, s], col_zk, opt_p['_tvp', k],
+                                           opt_x['_u', k, s_u], col_zk, opt_p['_tvp', k],
                                            opt_p['_p', current_scenario], _w)
 
                     # Add the collocation equations
@@ -1032,20 +1190,21 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
                     cons_lb.append(np.zeros((self.model.n_x, 1)))
                     cons_ub.append(np.zeros((self.model.n_x, 1)))
 
+                    k_eps = min(k, n_eps-1)
                     if self.nl_cons_check_colloc_points:
                         # Ensure nonlinear constraints on all collocation points
                         for i in range(n_total_coll_points):
                             nl_cons_k = self._nl_cons_fun(
-                                opt_x_unscaled['_x', k, s, i], opt_x_unscaled['_u', k, s], opt_x_unscaled['_z', k, s, i],
-                                opt_p['_tvp', k], opt_p['_p', current_scenario], opt_x_unscaled['_eps', k, s])
+                                opt_x_unscaled['_x', k, s, i], opt_x_unscaled['_u', k, s_u], opt_x_unscaled['_z', k, s, i],
+                                opt_p['_tvp', k], opt_p['_p', current_scenario], opt_x_unscaled['_eps', k_eps, s])
                             cons.append(nl_cons_k)
                             cons_lb.append(self._nl_cons_lb)
                             cons_ub.append(self._nl_cons_ub)
                     else:
                         # Ensure nonlinear constraints only on the beginning of the FE
                         nl_cons_k = self._nl_cons_fun(
-                            opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s], opt_x_unscaled['_z', k, s, 0],
-                            opt_p['_tvp', k], opt_p['_p', current_scenario], opt_x_unscaled['_eps', k, s])
+                            opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s_u], opt_x_unscaled['_z', k, s, 0],
+                            opt_p['_tvp', k], opt_p['_p', current_scenario], opt_x_unscaled['_eps', k_eps, s])
                         cons.append(nl_cons_k)
                         cons_lb.append(self._nl_cons_lb)
                         cons_ub.append(self._nl_cons_ub)
@@ -1054,24 +1213,31 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
                     # TODO: Add terminal constraints with an additional nl_cons
 
                     # Add contribution to the cost
-                    obj += omega[k] * self.lterm_fun(opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s],
+                    obj += omega[k] * self.lterm_fun(opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s_u],
                                                      opt_x_unscaled['_z', k, s, -1], opt_p['_tvp', k], opt_p['_p', current_scenario])
                     # Add slack variables to the cost
-                    obj += self.epsterm_fun(opt_x_unscaled['_eps', k, s])
+                    obj += self.epsterm_fun(opt_x_unscaled['_eps', k_eps, s])
 
                     # In the last step add the terminal cost too
                     if k == self.n_horizon - 1:
-                        obj += omega[k] * self.mterm_fun(opt_x_unscaled['_x', k + 1, s, -1], opt_p['_p', current_scenario])
+                        obj += omega[k] * self.mterm_fun(opt_x_unscaled['_x', k + 1, s, -1], opt_p['_tvp', k+1],
+                                                         opt_p['_p', current_scenario])
 
                     # U regularization:
                     if k == 0:
-                        obj += self.rterm_factor.cat.T@((opt_x['_u', 0, s]-opt_p['_u_prev']/self._u_scaling)**2)
+                        obj += self.rterm_factor.cat.T@((opt_x['_u', 0, s_u]-opt_p['_u_prev']/self._u_scaling)**2)
                     else:
-                        obj += self.rterm_factor.cat.T@((opt_x['_u', k, s]-opt_x['_u', k-1, parent_scenario[k][s]])**2)
+                        obj += self.rterm_factor.cat.T@((opt_x['_u', k, s_u]-opt_x['_u', k-1, parent_scenario[k][s_u]])**2)
 
                     # Calculate the auxiliary expressions for the current scenario:
                     opt_aux['_aux', k, s] = self.model._aux_expression_fun(
-                        opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s], opt_x_unscaled['_z', k, s, -1], opt_p['_tvp', k], opt_p['_p', current_scenario])
+                        opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s_u], opt_x_unscaled['_z', k, s, -1], opt_p['_tvp', k], opt_p['_p', current_scenario])
+
+                    # For some reason when working with MX, the "unused" aux values in the scenario tree must be set explicitly (they are not ever used...)
+                for s_ in range(n_scenarios[k],n_max_scenarios):
+                    opt_aux['_aux', k, s_] = self.model._aux_expression_fun(
+                        opt_x_unscaled['_x', k, s, -1], opt_x_unscaled['_u', k, s_u], opt_x_unscaled['_z', k, s, -1], opt_p['_tvp', k], opt_p['_p', current_scenario])
+
 
 
         if self.cons_check_colloc_points:   # Constraints for all collocation points.
@@ -1108,24 +1274,49 @@ class MPC(do_mpc.optimizer.Optimizer, do_mpc.model.IteratedVariables):
         self.ub_opt_x['_eps'] = self._eps_ub.cat
 
 
-        cons = vertcat(*cons)
-        self.cons_lb = vertcat(*cons_lb)
-        self.cons_ub = vertcat(*cons_ub)
+        # Write all created elements to self:
+        self._nlp_obj = obj
+        self._nlp_cons = cons
+        self._nlp_cons_lb = cons_lb
+        self._nlp_cons_ub = cons_ub
 
-        self.n_opt_lagr = cons.shape[0]
+
+        # Initialize copies of structures with numerical values (all zero):
+        self._opt_x_num = self._opt_x(0)
+        self.opt_x_num_unscaled = self._opt_x(0)
+        self._opt_p_num = self._opt_p(0)
+        self.opt_aux_num = self._opt_aux(0)
+
+        self.flags['prepare_nlp'] = True
+
+    def _create_nlp(self):
+        """Internal method. See detailed documentation in optimizer.create_nlp
+        """
+
+
+        self._nlp_cons = vertcat(*self._nlp_cons)
+        self._nlp_cons_lb = vertcat(*self._nlp_cons_lb)
+        self._nlp_cons_ub = vertcat(*self._nlp_cons_ub)
+
+        self.n_opt_lagr = self._nlp_cons.shape[0]
         # Create casadi optimization object:
         nlpsol_opts = {
             'expand': False,
             'ipopt.linear_solver': 'mumps',
         }.update(self.nlpsol_opts)
-        nlp = {'x': vertcat(opt_x), 'f': obj, 'g': cons, 'p': vertcat(opt_p)}
+        nlp = {'x': vertcat(self._opt_x), 'f': self._nlp_obj, 'g': self._nlp_cons, 'p': vertcat(self._opt_p)}
         self.S = nlpsol('S', 'ipopt', nlp, self.nlpsol_opts)
 
-        # Create copies of these structures with numerical values (all zero):
-        self.opt_x_num = self.opt_x(0)
-        self.opt_x_num_unscaled = self.opt_x(0)
-        self.opt_p_num = self.opt_p(0)
-        self.opt_aux_num = self.opt_aux(0)
 
         # Create function to caculate all auxiliary expressions:
-        self.opt_aux_expression_fun = Function('opt_aux_expression_fun', [opt_x, opt_p], [opt_aux])
+        self.opt_aux_expression_fun = Function('opt_aux_expression_fun', [self._opt_x, self._opt_p], [self._opt_aux])
+
+
+        # Gather meta information:
+        meta_data = {key: getattr(self, key) for key in self.data_fields}
+        meta_data.update({'structure_scenario': self.scenario_tree['structure_scenario']})
+        self.data.set_meta(**meta_data)
+
+        self._prepare_data()
+
+        self.flags['setup'] = True
