@@ -45,7 +45,7 @@ class Estimator(do_mpc.model.IteratedVariables):
         self.model = model
         do_mpc.model.IteratedVariables.__init__(self)
 
-        assert model.flags['setup'] == True, 'Model for estimator was not setup. After the complete model creation call model.setup_model().'
+        assert model.flags['setup'] == True, 'Model for estimator was not setup. After the complete model creation call model.setup().'
 
         self.data = do_mpc.data.Data(model)
         self.data.dtype = 'Estimator'
@@ -125,9 +125,58 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
 
     6. Optionally, set further (non-linear) constraints with :py:func:`set_nl_cons`.
 
-    7. Use :py:func:`get_p_template` and :py:func:`set_p_fun` to set the function for the parameters.
+    7. Use :py:func:`get_p_template` and :py:func:`set_p_fun` to set the function for the (not estimated) parameters.
 
-    8. Finally, call :py:func:`setup`.
+    8. Use :py:meth:`get_tvp_template` and :py:meth:`set_tvp_fun` to create a method to obtain new time-varying parameters at each iteration.
+
+    9. To finalize the class configuration there are two routes. The default approach is to call :py:meth:`setup`. For deep customization use the combination of :py:meth:`prepare_nlp` and :py:meth:`create_nlp`. See graph below for an illustration of the process.
+
+
+    .. graphviz::
+        :name: route_to_setup
+        :caption: Route to setting up the MHE class.
+        :align: center
+
+        digraph G {
+            graph [fontname = "helvetica"];
+            rankdir=LR;
+
+            subgraph cluster_main {
+                node [fontname = "helvetica", shape=box, fontcolor="#404040", color="#707070"];
+                edge [fontname = "helvetica", color="#707070"];
+
+                start [label="Two ways to setup"];
+                setup [label="setup", href="../api/do_mpc.estimator.MHE.setup.html", target="_top", fontname = "Consolas"];
+                create_nlp [label="create_nlp", href="../api/do_mpc.estimator.MHE.create_nlp.html", target="_top", fontname = "Consolas"];
+                process [label="Modify NLP"];
+                prepare_nlp [label="prepare_nlp", href="../api/do_mpc.estimator.MHE.prepare_nlp.html", target="_top", fontname = "Consolas"];
+                finish [label="Configured MHE class"]
+                start -> setup, prepare_nlp;
+                prepare_nlp -> process;
+                process -> create_nlp;
+                setup, create_nlp -> finish;
+                color=none;
+            }
+
+            subgraph cluster_modification {
+                rankdir=TB;
+                node [fontname = "helvetica", shape=box, fontcolor="#404040", color="#707070"];
+                edge [fontname = "helvetica", color="#707070"];
+                opt_x [label="opt_x", href="../api/do_mpc.estimator.MHE.opt_x.html", target="_top", fontname = "Consolas"];
+                opt_p [label="opt_p", href="../api/do_mpc.estimator.MHE.opt_p.html", target="_top", fontname = "Consolas"];
+                nlp_cons [label="nlp_cons", href="../api/do_mpc.estimator.MHE.nlp_cons.html", target="_top", fontname = "Consolas"];
+                nlp_obj [label="nlp_obj", href="../api/do_mpc.estimator.MHE.nlp_obj.html", target="_top", fontname = "Consolas"];
+
+                opt_x -> nlp_cons, nlp_obj;
+                opt_p -> nlp_cons, nlp_obj;
+
+                label = "Attributes to modify the NLP.";
+		        color=black;
+            }
+
+            nlp_cons -> process;
+            nlp_obj -> process;
+        }
 
     .. warning::
 
@@ -165,6 +214,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
             'collocation_deg',
             'collocation_ni',
             'nl_cons_check_colloc_points',
+            'nl_cons_single_slack',
             'cons_check_colloc_points',
             'store_full_solution',
             'store_lagr_multiplier',
@@ -179,6 +229,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         self.collocation_deg = 2
         self.collocation_ni = 1
         self.nl_cons_check_colloc_points = False
+        self.nl_cons_single_slack = False
         self.cons_check_colloc_points = True
         self.store_full_solution = False
         self.store_lagr_multiplier = True
@@ -192,15 +243,29 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
 
         # Create seperate structs for the estimated and the set parameters (the union of both are all parameters of the model.)
         _p = model._p
-        self._p_est  = struct_symSX(
+        self._p_est  = self.model.sv.sym_struct(
             [entry('default', shape=(0,0))]+
-            [entry(p_i, sym=_p[p_i]) for p_i in _p.keys() if p_i in p_est_list]
+            [entry(p_i, shape=_p[p_i].shape) for p_i in _p.keys() if p_i in p_est_list]
         )
-        self._p_set  = struct_symSX(
-            [entry(p_i, sym=_p[p_i]) for p_i in _p.keys() if p_i not in p_est_list]
+        self._p_set  = self.model.sv.sym_struct(
+            [entry(p_i, shape=_p[p_i].shape) for p_i in _p.keys() if p_i not in p_est_list]
         )
+
+        # Enable to "unite" _p_est and _p_set to _p
+        p_cat = vertcat(_p)
+        _p_subs = []
+        for name in _p.keys():
+            if name in self._p_est.keys():
+                _p_subs.append(self._p_est[name])
+            elif name in self._p_set.keys():
+                _p_subs.append(self._p_set[name])
+
+        # In the expression p_cat substitute all variables from _p with the elements from _p_est and _p_set:
+        p_cat = substitute(p_cat, _p, vertcat(*_p_subs))
+
         # Function to obtain full set of parameters from the seperate structs (while obeying the order):
-        self._p_cat_fun = Function('p_cat_fun', [self._p_est, self._p_set], [_p])
+        self._p_cat_fun = Function('p_cat_fun', [self._p_est, self._p_set], [p_cat])
+
 
         self.n_p_est = self._p_est.shape[0]
         self.n_p_set = self._p_set.shape[0]
@@ -226,20 +291,20 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         self._x = self.model._x
 
         self._p_est_prev = copy.copy(self._p_est)
-        self._p_est = self._p_est
+        self._p = self.model._p
 
         self._w = self.model._w
         self._v = self.model._v
 
         # Flags are checked when calling .setup.
-        self.flags = {
+        self.flags.update({
             'setup': False,
             'set_tvp_fun': False,
             'set_p_fun': False,
             'set_y_fun': False,
             'set_objective': False,
             'set_initial_guess': False,
-        }
+        })
 
     @property
     def p_est0(self):
@@ -381,6 +446,89 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
     def opt_p_num(self, val):
         self._opt_p_num = val
 
+    @property
+    def opt_x(self):
+        """Full structure of the (symbolic) MHE optimization variables.
+
+        The attribute is a CasADi numeric structure with nested power indices.
+        It can be indexed as follows:
+
+        ::
+
+            # dynamic states:
+            opt_x['_x', time_step, collocation_point, _x_name]
+            # algebraic states:
+            opt_x['_z', time_step, collocation_point, _z_name]
+            # inputs:
+            opt_x['_u', time_step, _u_name]
+            # estimated parameters:
+            opt_x_Num['_p_est', _p_names]
+            # slack variables for soft constraints:
+            opt_x['_eps', time_step, _nl_cons_name]
+
+        The names refer to those given in the :py:class:`do_mpc.model.Model` configuration.
+        Further indices are possible, if the variables are itself vectors or matrices.
+
+        The attribute can be used to alter the objective function or constraints of the NLP.
+
+        .. note::
+
+            The attribute ``opt_x`` carries the scaled values of all variables.
+
+        .. warning::
+
+            Do not tweak or overwrite this attribute unless you known what you are doing.
+
+        .. note::
+
+            The attribute is populated when calling :py:func:`setup` or :py:func:`prepare_nlp`
+
+        """
+        return self._opt_x
+
+    @opt_x.setter
+    def opt_x(self, val):
+        self._opt_x = val
+
+    @property
+    def opt_p(self):
+        """Full structure of (symbolic) MHE parameters.
+
+        The attribute can be used to alter the objective function or constraints of the NLP.
+
+        The attribute is a CasADi numeric structure with nested power indices.
+        It can be indexed as follows:
+
+        ::
+
+            # previously estimated state:
+            opt_p['_x_prev', _x_name]
+            # previously estimated parameters:
+            opt_p['_p_est_prev', _x_name]
+            # known parameters
+            opt_p['_p_set', _p_name]
+            # time-varying parameters:
+            opt_p['_tvp', time_step, _tvp_name]
+            # sequence of measurements:
+            opt_p['_y_meas', time_step, _y_name]
+
+        The names refer to those given in the :py:class:`do_mpc.model.Model` configuration.
+        Further indices are possible, if the variables are itself vectors or matrices.
+
+        .. warning::
+
+            Do not tweak or overwrite this attribute unless you known what you are doing.
+
+        .. note::
+
+            The attribute is populated when calling :py:func:`setup` or :py:func:`create_nlp`.
+
+        """
+        return self._opt_p
+
+    @opt_p.setter
+    def opt_p(self, val):
+        self._opt_p = val
 
     def set_param(self, **kwargs):
         """Method to set the parameters of the :py:class:`MHE` class. Parameters must be passed as pairs of valid keywords and respective argument.
@@ -419,7 +567,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         :param meas_from_data: Default option to retrieve past measurements for the MHE optimization problem. The :py:func:`set_y_fun` is called during setup.
         :type meas_from_data: bool
 
-        :param state_discretization: Choose the state discretization for continuous models. Currently only ``'collocation'`` is available. Defaults to ``'collocation'``.
+        :param state_discretization: Choose the state discretization for continuous models. Currently only ``'collocation'`` is available. Defaults to ``'collocation'``. Has no effect if model is created in ``discrete`` type.
         :type state_discretization: str
 
         :param collocation_type: Choose the collocation type for continuous models with collocation as state discretization. Currently only ``'radau'`` is available. Defaults to ``'radau'``.
@@ -437,6 +585,9 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         :param cons_check_colloc_points: For orthogonal collocation choose whether the linear bounds set with :py:attr:`bounds` are evaluated once per finite Element or for each collocation point. Defaults to ``True`` (for all collocation points).
         :type cons_check_colloc_points: bool
 
+        :param nl_cons_single_slack: If ``True``, soft-constraints set with :py:func:`set_nl_cons` introduce only a single slack variable for the entire horizon. Defaults to ``False``.
+        :type nl_cons_single_slack: bool
+
         :param store_full_solution: Choose whether to store the full solution of the optimization problem. This is required for animating the predictions in post processing. However, it drastically increases the required storage. Defaults to False.
         :type store_full_solution: bool
 
@@ -454,12 +605,12 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
             ::
 
                 optimizer.set_param(nlpsol_opts = {'ipopt.linear_solver': 'MA27'})
-        .. note:: To surpress the output of IPOPT, please use:
+        .. note:: To suppress the output of IPOPT, please use:
 
             ::
 
-                surpress_ipopt = {'ipopt.print_level':0, 'ipopt.sb': 'yes', 'print_time':0}
-                optimizer.set_param(nlpsol_opts = surpress_ipopt)
+                suppress_ipopt = {'ipopt.print_level':0, 'ipopt.sb': 'yes', 'print_time':0}
+                optimizer.set_param(nlpsol_opts = suppress_ipopt)
 
         """
         assert self.flags['setup'] == False, 'Setting parameters after setup is prohibited.'
@@ -556,16 +707,37 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         assert arrival_cost.shape == (1,1), 'arrival_cost must have shape=(1,1). You have {}'.format(arrival_cost.shape)
         assert self.flags['setup'] == False, 'Cannot call .set_objective after .setup.'
 
+        # Replace model symbolic variables self.model._p with the new variables self._p_est and self._p_set:
+        _p = self._p_cat_fun(self._p_est, self._p_set)
+
+        arrival_cost = substitute(arrival_cost, self.model._p, _p)
+        #stage_cost = substitute(stage_cost, self.model._p, _p)
+
+        #arrival_cost = substitute(arrival_cost, self._p_est, vertcat(*[self.model._p[name] for name in self._p_est.keys()]))
+        #arrival_cost = substitute(arrival_cost, self._p_set, vertcat(*[self.model._p[name] for name in self._p_set.keys()]))
+
+        stage_cost = substitute(stage_cost, self._p_est, vertcat(*[self.model._p[name] for name in self._p_est.keys()]))
+        stage_cost = substitute(stage_cost, self._p_set, vertcat(*[self.model._p[name] for name in self._p_set.keys()]))
+
 
         stage_cost_input = self._w, self._v, self.model._tvp, self.model._p
-        err_msg = 'objective cost equation must be solely depending on w, v, p and tvp.'
-        assert set(symvar(stage_cost)).issubset(set(symvar(vertcat(*stage_cost_input)))), err_msg
         self.stage_cost_fun = Function('stage_cost_fun', [*stage_cost_input], [stage_cost])
 
         arrival_cost_input = self._x, self._x_prev, self._p_est, self._p_est_prev, self._p_set
-        err_msg = 'Arrival cost equation must be solely depending on x_0, x_prev, p_0, p_prev, p_set'
-        assert set(symvar(arrival_cost)).issubset(set(symvar(vertcat(*arrival_cost_input)))), err_msg
-        self.arrival_cost_fun = Function('arrival_cost_fun', arrival_cost_input, [arrival_cost])
+        self.arrival_cost_fun = Function('arrival_cost_fun', [*arrival_cost_input], [arrival_cost])
+
+        # Check if stage_cost_fun and arrival_cost_fun use invalid variables as inputs.
+        # For the check we evaluate the function with dummy inputs and expect a DM output.
+        try:
+            self.stage_cost_fun(*[input_i(0) for input_i in stage_cost_input])
+        except:
+            err_msg = 'objective cost equation must be solely depending on w, v, p and tvp.'
+            raise Exception(err_msg)
+        try:
+            self.arrival_cost_fun(*[input_i(0) for input_i in arrival_cost_input])
+        except:
+            err_msg = 'Arrival cost equation must be solely depending on x_0, x_prev, p_0, p_prev, p_set'
+            raise Exception(err_msg)
 
         self.flags['set_objective'] = True
 
@@ -632,10 +804,10 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         :type P_w: numpy.ndarray, casadi.SX, casadi.DM
         """
 
-        input_types = (np.ndarray, casadi.SX, casadi.DM)
+        input_types = (np.ndarray, casadi.SX, casadi.MX, casadi.DM)
         err_msg = '{name} must be of type {type_set}, you have {type_is}'
         assert isinstance(P_x, input_types), err_msg.format(name='P_x', type_set = input_types, type_is = type(P_x))
-        input_types = (np.ndarray, casadi.SX, casadi.DM, type(None))
+        input_types = (np.ndarray, casadi.SX, casadi.MX, casadi.DM, type(None))
         assert isinstance(P_v, input_types), err_msg.format(name='P_v', type_set = input_types, type_is = type(P_v))
         assert isinstance(P_p, input_types), err_msg.format(name='P_p', type_set = input_types, type_is = type(P_p))
         assert isinstance(P_w, input_types), err_msg.format(name='P_w', type_set = input_types, type_is = type(P_w))
@@ -686,6 +858,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
             arrival_cost += dp.T@P_p@dp
 
 
+
         # Set MHE objective:
         self.set_objective(stage_cost, arrival_cost)
 
@@ -761,7 +934,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         :return: y_template
         :rtype: struct_symSX
         """
-        y_template = struct_symSX([
+        y_template = self.model.sv.sym_struct([
             entry('y_meas', repeat=self.n_horizon, struct=self._y_meas)
         ])
         return y_template(0)
@@ -869,20 +1042,10 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
 
             After this call, the :py:func:`solve` and :py:func:`make_step` method is applicable.
         """
-        self._setup_nl_cons()
 
-        # Concatenate _p_est_scaling und _p_set_scaling to p_scaling (and make it a struct again)
-        self._p_scaling = self.model._p(self._p_cat_fun(self._p_est_scaling, self._p_set_scaling))
+        self.prepare_nlp()
+        self.create_nlp()
 
-        # Gather meta information:
-        meta_data = {key: getattr(self, key) for key in self.data_fields}
-        self.data.set_meta(**meta_data)
-
-        self._check_validity()
-        self._setup_mhe_optim_problem()
-
-        self._prepare_data()
-        self.flags['setup'] = True
 
     def make_step(self, y0):
         """Main method of the class during runtime. This method is called at each timestep
@@ -911,7 +1074,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         elif isinstance(y0, structure3.DMStruct):
             y0 = y0.cat
         else:
-            raise Exception('Invalid type {} for y0. Must be {}'.format(type(y0), (np.ndarray, casadi.DM, structre3.DMStruct)))
+            raise Exception('Invalid type {} for y0. Must be {}'.format(type(y0), (np.ndarray, casadi.DM, structure3.DMStruct)))
 
         # Check input shape.
         n_val = np.prod(y0.shape)
@@ -985,27 +1148,41 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
 
         return x_next.full()
 
-    def _setup_mhe_optim_problem(self):
-        """Private method of the MHE class to construct the MHE optimization problem.
-        The method depends on inherited methods from the :py:class:`do_mpc.optimizer.Optimizer`,
-        such as :py:func:`do_mpc.optimizer.Optimizer._setup_discretization` and
-        :py:func:`do_mpc.optimizer.Optimizer._setup_scenario_tree`.
-
-        The MPC has a similar method with similar structure.
+    def _prepare_nlp(self):
+        """Internal method. See detailed documentation in optimizer.prepare_nlp
         """
+
+        nl_cons_input = self.model['x', 'u', 'z', 'tvp']
+        nl_cons_input += [self._p_est, self._p_set]
+        self._setup_nl_cons(nl_cons_input)
+        self._check_validity()
+
+        # Concatenate _p_est_scaling und _p_set_scaling to p_scaling (and make it a struct again)
+        self._p_scaling = self.model._p(self._p_cat_fun(self._p_est_scaling, self._p_set_scaling))
+
+
         # Obtain an integrator (collocation, discrete-time) and the amount of intermediate (collocation) points
         ifcn, n_total_coll_points = self._setup_discretization()
+
+        # How many slack variables (for soft constraints) are introduced over the horizon.
+        if self.nl_cons_single_slack:
+            n_eps = 1
+        else:
+            n_eps = self.n_horizon
+
+
         # Create struct for optimization variables:
-        self.opt_x = opt_x = struct_symSX([
+        self._opt_x = opt_x = self.model.sv.sym_struct([
             entry('_x', repeat=[self.n_horizon+1, 1+n_total_coll_points], struct=self.model._x),
             entry('_z', repeat=[self.n_horizon,   max(n_total_coll_points,1)], struct=self.model._z),
             entry('_u', repeat=[self.n_horizon], struct=self.model._u),
             entry('_w', repeat=[self.n_horizon], struct=self.model._w),
             entry('_v', repeat=[self.n_horizon], struct=self.model._v),
-            entry('_eps', repeat=[self.n_horizon], struct=self._eps),
+            entry('_eps', repeat=[n_eps], struct=self._eps),
             entry('_p_est', struct=self._p_est),
         ])
-        self.n_opt_x = self.opt_x.shape[0]
+
+        self.n_opt_x = opt_x.shape[0]
         # NOTE: The entry _x[k,:] starts with the collocation points from s to b at time k
         #       and the last point contains the child node
         # NOTE: Currently there exist dummy collocation points for the initial state (for each branch)
@@ -1022,7 +1199,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
 
 
         # Create struct for optimization parameters:
-        self.opt_p = opt_p = struct_symSX([
+        self._opt_p = opt_p = self.model.sv.sym_struct([
             entry('_x_prev', struct=self.model._x),
             entry('_p_est_prev', struct=self._p_est_prev),
             entry('_p_set', struct=self._p_set),
@@ -1032,13 +1209,13 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         self.n_opt_p = opt_p.shape[0]
 
         # Dummy struct with symbolic variables
-        self.aux_struct = struct_symSX([
+        self.aux_struct = self.model.sv.sym_struct([
             entry('_aux', repeat=[self.n_horizon], struct=self.model._aux_expression)
         ])
         # Create mutable symbolic expression from the struct defined above.
-        self.opt_aux = opt_aux = struct_SX(self.aux_struct)
+        self._opt_aux = opt_aux = self.model.sv.struct(self.aux_struct)
 
-        self.n_opt_aux = self.opt_aux.shape[0]
+        self.n_opt_aux = opt_aux.shape[0]
 
         self.lb_opt_x = opt_x(-np.inf)
         self.ub_opt_x = opt_x(np.inf)
@@ -1061,7 +1238,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         obj += arrival_cost
 
         # Get concatenated parameters vector containing the estimated and fixed parameters (scaled)
-        _p = self._p_cat_fun(self.opt_x['_p_est'], self.opt_p['_p_set']/self._p_set_scaling)
+        _p = self._p_cat_fun(opt_x['_p_est'], opt_p['_p_set']/self._p_set_scaling)
 
         # For all control intervals
         for k in range(self.n_horizon):
@@ -1091,12 +1268,13 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
             cons_lb.append(np.zeros((self.model.n_y, 1)))
             cons_ub.append(np.zeros((self.model.n_y, 1)))
 
+            k_eps = min(k, n_eps-1)
             if self.nl_cons_check_colloc_points:
                 # Ensure nonlinear constraints on all collocation points
                 for i in range(n_total_coll_points):
                     nl_cons_k = self._nl_cons_fun(
                         opt_x_unscaled['_x', k, i], opt_x_unscaled['_u', k], opt_x_unscaled['_z', k, i],
-                        opt_p['_tvp', k], _p, opt_x_unscaled['_eps', k])
+                        opt_p['_tvp', k], opt_x['_p_est'], opt_p['_p_set'], opt_x_unscaled['_eps', k_eps])
                     cons.append(nl_cons_k)
                     cons_lb.append(self._nl_cons_lb)
                     cons_ub.append(self._nl_cons_ub)
@@ -1104,7 +1282,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
                 # Ensure nonlinear constraints only on the beginning of the FE
                 nl_cons_k = self._nl_cons_fun(
                     opt_x_unscaled['_x', k, -1], opt_x_unscaled['_u', k], opt_x_unscaled['_z', k, 0],
-                    opt_p['_tvp', k], _p, opt_x_unscaled['_eps', k])
+                    opt_p['_tvp', k], opt_x['_p_est'], opt_p['_p_set'], opt_x_unscaled['_eps', k_eps])
                 cons.append(nl_cons_k)
                 cons_lb.append(self._nl_cons_lb)
                 cons_ub.append(self._nl_cons_ub)
@@ -1119,7 +1297,7 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
             )
 
             # Add slack variables to the cost
-            obj += self.epsterm_fun(opt_x_unscaled['_eps', k])
+            obj += self.epsterm_fun(opt_x_unscaled['_eps', k_eps])
 
 
             # Calculate the auxiliary expressions for the current scenario:
@@ -1156,24 +1334,64 @@ class MHE(do_mpc.optimizer.Optimizer, Estimator):
         self.lb_opt_x['_p_est'] = self._p_est_lb.cat/self._p_est_scaling
         self.ub_opt_x['_p_est'] = self._p_est_ub.cat/self._p_est_scaling
 
-        cons = vertcat(*cons)
-        self.cons_lb = vertcat(*cons_lb)
-        self.cons_ub = vertcat(*cons_ub)
 
-        self.n_opt_lagr = cons.shape[0]
+        # Write all created elements to self:
+        self._nlp_obj = obj
+        self._nlp_cons = cons
+        self._nlp_cons_lb = cons_lb
+        self._nlp_cons_ub = cons_ub
+
+        # Initialize copies of structures with numerical values (all zero):
+        self._opt_x_num = self._opt_x(0)
+        self.opt_x_num_unscaled = self._opt_x(0)
+        self._opt_p_num = self._opt_p(0)
+        self.opt_aux_num = self._opt_aux(0)
+
+        self.flags['prepare_nlp'] = True
+
+
+    def _create_nlp(self):
+        """Internal method. See detailed documentation in optimizer.create_nlp
+        """
+
+
+        self._nlp_cons = vertcat(*self._nlp_cons)
+        self._nlp_cons_lb = vertcat(*self._nlp_cons_lb)
+        self._nlp_cons_ub = vertcat(*self._nlp_cons_ub)
+
+
+        # Validity check:
+        _test_obj_fun = Function('f', [self._opt_x, self._opt_p], [self._nlp_obj])
+        _test_cons_fun = Function('f', [self._opt_x, self._opt_p], [self._nlp_cons])
+        try:
+            _test_obj_fun(self.opt_x_num,self.opt_p_num)
+        except:
+            err_msg = 'The MHE optimization problem objective function contains unknown symbolic variables.'
+            raise Exception(err_msg)
+        try:
+            _test_cons_fun(self.opt_x_num,self.opt_p_num)
+        except:
+            err_msg = 'The MHE optimization problem constraint function contains unknown symbolic variables.'
+            raise Exception(err_msg)
+
+
+        self.n_opt_lagr = self._nlp_cons.shape[0]
         # Create casadi optimization object:
         nlpsol_opts = {
             'expand': False,
             'ipopt.linear_solver': 'mumps',
         }.update(self.nlpsol_opts)
-        nlp = {'x': vertcat(opt_x), 'f': obj, 'g': cons, 'p': vertcat(opt_p)}
+        nlp = {'x': vertcat(self._opt_x), 'f': self._nlp_obj, 'g': self._nlp_cons, 'p': vertcat(self._opt_p)}
         self.S = nlpsol('S', 'ipopt', nlp, self.nlpsol_opts)
 
-        # Create copies of these structures with numerical values (all zero):
-        self.opt_x_num = self.opt_x(0)
-        self.opt_x_num_unscaled = self.opt_x(0)
-        self.opt_p_num = self.opt_p(0)
-        self.opt_aux_num = self.opt_aux(0)
+
 
         # Create function to caculate all auxiliary expressions:
-        self.opt_aux_expression_fun = Function('opt_aux_expression_fun', [opt_x, opt_p], [opt_aux])
+        self.opt_aux_expression_fun = Function('opt_aux_expression_fun', [self._opt_x, self._opt_p], [self._opt_aux])
+
+        # Gather meta information:
+        meta_data = {key: getattr(self, key) for key in self.data_fields}
+        self.data.set_meta(**meta_data)
+
+        self._prepare_data()
+        self.flags['setup'] = True
