@@ -49,6 +49,11 @@ class Optimizer:
         # with the default numerical value.
         # This returns an identical numerical structure with all values set to the passed value.
 
+        self.flags = {
+            'prepare_nlp': False,
+            'setup': False,
+        }
+
         self._x_lb = self.model._x(-np.inf)
         self._x_ub = self.model._x(np.inf)
 
@@ -65,13 +70,160 @@ class Optimizer:
 
         # Lists for further non-linear constraints (optional). Constraints are formulated as cons < ub
         self.nl_cons_list = [
-            {'expr_name': 'default', 'expr': DM(), 'ub': DM()}
+            {'expr_name': 'default', 'expr': DM([]), 'ub': DM([])}
         ]
         self.slack_vars_list = [
-            {'slack_name': 'default', 'var':SX.sym('default',(0,0)), 'ub': DM()}
+            {'slack_name': 'default', 'shape':0, 'ub': DM([]), 'penalty': 0}
         ]
         self.slack_cost = 0
 
+    @property
+    def nlp_obj(self):
+        """Query and modify (symbolically) the NLP objective function.
+        Use the variables in :py:attr:`opt_x` and :py:attr:`opt_p`.
+
+        It is advised to add to the current objective, e.g.:
+
+        ::
+
+            mpc.prepare_nlp()
+            # Modify the objective
+            mpc.nlp_obj += sum1(vertcat(*mpc.opt_x['_x', -1, 0])**2)
+            # Finish creating the NLP
+            mpc.create_nlp()
+
+        See the documentation of :py:attr:`opt_x` and :py:attr:`opt_p` on how to query these attributes.
+
+        .. warning::
+
+            This is a VERY low level feature and should be used with extreme caution.
+            It is easy to break the code.
+
+            Be especially careful NOT to accidentially overwrite the default objective.
+
+        .. note::
+
+            Modifications must be done after calling :py:meth:`prepare_nlp`
+            and before calling :py:meth:`create_nlp`
+
+        """
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling MPC.prepare_nlp or MPC.setup'
+        return self._nlp_obj
+
+    @nlp_obj.setter
+    def nlp_obj(self, val):
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling MPC.prepare_nlp or MPC.setup'
+        assert not self.flags['setup'], 'Cannot change attribute after calling MPC.create_nlp or MPC.setup'
+        self._nlp_obj = val
+
+    @property
+    def nlp_cons(self):
+        """Query and modify (symbolically) the NLP constraints.
+        Use the variables in :py:attr:`opt_x` and :py:attr:`opt_p`.
+
+        Prior to calling :py:meth:`create_nlp` this attribute returns a list of symbolic constraints.
+        After calling :py:meth:`create_nlp` this attribute returns the concatenation of this list
+        and the attribute cannot be altered anymore.
+
+        It is advised to append to the current list of :py:attr:`nlp_cons`:
+
+        ::
+
+            mpc.prepare_nlp()
+
+            # Create new constraint: Input at timestep 0 and 1 must be identical.
+            extra_cons = mpc.opt_x['_u', 0, 0]-mpc.opt_x['_u',1, 0]
+            mpc.nlp_cons.append(
+                extra_cons
+            )
+
+            # Create appropriate upper and lower bound (here they are both 0 to create an equality constraint)
+            mpc.nlp_cons_lb.append(np.zeros(extra_cons.shape))
+            mpc.nlp_cons_ub.append(np.zeros(extra_cons.shape))
+
+            mpc.create_nlp()
+
+        See the documentation of :py:attr:`opt_x` and :py:attr:`opt_p` on how to query these attributes.
+
+        .. warning::
+
+            This is a VERY low level feature and should be used with extreme caution.
+            It is easy to break the code.
+
+            Be especially careful NOT to accidentially overwrite the default objective.
+
+        .. note::
+
+            Modifications must be done after calling :py:meth:`prepare_nlp`
+            and before calling :py:meth:`create_nlp`
+
+        """
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling MPC.prepare_nlp or MPC.setup'
+        return self._nlp_cons
+
+    @nlp_cons.setter
+    def nlp_cons(self, val):
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling prepare_nlp or setup'
+        assert not self.flags['setup'], 'Cannot change attribute after calling create_nlp or setup'
+        self._nlp_cons = val
+
+
+    @property
+    def nlp_cons_lb(self):
+        """Query and modify the lower bounds of the :py:attr:`nlp_cons`.
+
+        Prior to calling :py:meth:`create_nlp` this attribute returns a list of lower bounds
+        matching the list of constraints obtained with :py:attr:`nlp_cons`.
+        After calling :py:meth:`create_nlp` this attribute returns the concatenation of this list.
+
+        Values for lower (and upper) bounds MUST be added when adding new constraints to :py:attr:`nlp_cons`.
+
+        .. warning::
+
+            This is a VERY low level feature and should be used with extreme caution.
+            It is easy to break the code.
+
+        .. note::
+
+            Modifications must be done after calling :py:meth:`prepare_nlp`
+
+        """
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling MPC.prepare_nlp or MPC.setup'
+        return self._nlp_cons_lb
+
+    @nlp_cons_lb.setter
+    def nlp_cons_lb(self, val):
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling prepare_nlp or setup'
+        self._nlp_cons_lb = val
+
+    @property
+    def nlp_cons_ub(self):
+        """Query and modify the upper bounds of the :py:attr:`nlp_cons`.
+
+        Prior to calling :py:meth:`create_nlp` this attribute returns a list of upper bounds
+        matching the list of constraints obtained with :py:attr:`nlp_cons`.
+        After calling :py:meth:`create_nlp` this attribute returns the concatenation of this list.
+
+        Values for upper (and lower) bounds MUST be added when adding new constraints to :py:attr:`nlp_cons`.
+
+        .. warning::
+
+            This is a VERY low level feature and should be used with extreme caution.
+            It is easy to break the code.
+
+        .. note::
+
+            Modifications must be done after calling :py:meth:`prepare_nlp`
+
+        """
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling MPC.prepare_nlp or MPC.setup'
+        return self._nlp_cons_ub
+
+    @nlp_cons_ub.setter
+    def nlp_cons_ub(self, val):
+        assert self.flags['prepare_nlp'], 'Cannot query attribute prior to calling prepare_nlp or setup'
+        self._nlp_cons_ub = val
+
 
     @IndexedProperty
     def bounds(self, ind):
@@ -123,7 +275,7 @@ class Optimizer:
         var_struct = getattr(self, query)
 
         err_msg = 'Calling .bounds with {} is not valid. Possible keys are {}.'
-        assert (var_name[0] if isinstance(var_name, tuple) else var_name) in var_struct.keys(), msg.format(ind, var_struct.keys())
+        assert (var_name[0] if isinstance(var_name, tuple) else var_name) in var_struct.keys(), err_msg.format(ind, var_struct.keys())
 
         return var_struct[var_name]
 
@@ -315,61 +467,63 @@ class Optimizer:
         :raises assertion: expr must be a casadi SX or MX type
 
         :return: Returns the newly created expression. Expression can be used e.g. for the RHS.
-        :rtype: casadi.SX
+        :rtype: casadi.SX or casadi.MX
         """
-        assert self.flags['setup'] == False, 'Cannot call .set_expression after .setup_model.'
+        assert self.flags['setup'] == False, 'Cannot call .set_expression after .setup().'
         assert isinstance(expr_name, str), 'expr_name must be str, you have: {}'.format(type(expr_name))
         assert isinstance(expr, (casadi.SX, casadi.MX)), 'expr must be a casadi SX or MX type, you have: {}'.format(type(expr))
         assert isinstance(ub, (int, float, np.ndarray)), 'ub must be float, int or numpy.ndarray, you have: {}'.format(type(ub))
         assert isinstance(soft_constraint, bool), 'soft_constraint must be boolean, you have: {}'.format(type(soft_constraint))
 
         if soft_constraint==True:
-            # Introduce new slack variable:
-            epsilon = SX.sym('eps_'+expr_name,*expr.shape)
-            # Change expression
-            expr = expr-epsilon
             # Add slack variable to list of slack variables:
             self.slack_vars_list.extend([
-                {'slack_name': expr_name, 'var': epsilon, 'ub': maximum_violation}
+                {'slack_name': expr_name, 'shape': expr.shape, 'ub': maximum_violation, 'penalty': penalty_term_cons}
             ])
-            # Add cost contribution:
-            self.slack_cost += sum1(penalty_term_cons*epsilon)
 
+        # All operations to make the soft constraints work are in _setup_nl_cons.
 
         self.nl_cons_list.extend([
             {'expr_name': expr_name, 'expr': expr, 'ub' : ub}])
 
         return expr
 
-    def _setup_nl_cons(self):
+    def _setup_nl_cons(self, nl_cons_input):
         """Private method that is called from :py:func:`do_mpc.controller.MPC.setup` or :py:func:`do_mpc.estimator.MHE.setup`.
         Afterwards no further non-linear constraints can be added with the :py:func:`Optimizer.set_nl_cons` method.
 
         This is not part of the public API. Do not call this method.
+
+        ``nl_cons_input`` list of symbolic variables used as input to the nl_cons function.
         """
         # Create struct for soft constraints:
-        self._eps = _eps = struct_symSX([
-            entry(slack_i['slack_name'], sym=slack_i['var']) for slack_i in self.slack_vars_list
+        self._eps = _eps = self.model.sv.sym_struct([
+            entry(slack_i['slack_name'], shape=slack_i['shape']) for slack_i in self.slack_vars_list
+        ])
+        # Create struct for _nl_cons:
+        # Use the previously defined sym variables to declare shape and symbolic variable.
+        self._nl_cons = self.model.sv.struct([
+            entry(expr_i['expr_name'], expr=expr_i['expr']) for expr_i in self.nl_cons_list
         ])
-        self.n_eps = _eps.shape[0]
 
+        self.n_eps = _eps.shape[0]
         # Create bounds:
         self._eps_lb = _eps(0.0)
         self._eps_ub = _eps(np.inf)
-        # Set bounds:
+
+        # Set bounds, add slack variable to constraint and add cost.
         for slack_i in self.slack_vars_list:
             self._eps_ub[slack_i['slack_name']] = slack_i['ub']
+            self._nl_cons[slack_i['slack_name']] -= self._eps[slack_i['slack_name']]
+            self.slack_cost += sum1(slack_i['penalty']*self._eps[slack_i['slack_name']])
+
         # Objective function epsilon contribution:
         self.epsterm_fun = Function('epsterm', [_eps], [self.slack_cost])
 
-        # Create struct for _nl_cons:
-        # Use the previously defined SX.sym variables to declare shape and symbolic variable.
-        self._nl_cons = struct_SX([
-            entry(expr_i['expr_name'], expr=expr_i['expr']) for expr_i in self.nl_cons_list
-        ])
         # Make function from these expressions:
-        _x, _u, _z, _tvp, _p = self.model['x', 'u', 'z', 'tvp', 'p']
-        self._nl_cons_fun = Function('nl_cons_fun', [_x, _u, _z, _tvp, _p, _eps], [self._nl_cons])
+        nl_cons_input += [_eps]
+        self._nl_cons_fun = Function('nl_cons_fun', nl_cons_input, [self._nl_cons])
+
         # Create bounds:
         self._nl_cons_ub = self._nl_cons(np.inf)
         self._nl_cons_lb = self._nl_cons(-np.inf)
@@ -381,7 +535,7 @@ class Optimizer:
     def get_tvp_template(self):
         """Obtain output template for :py:func:`set_tvp_fun`.
 
-        The method returns a structured object with ``n_horizon`` elements,
+        The method returns a structured object with ``n_horizon+1`` elements,
         and a set of time-varying parameters (as defined in :py:class:`do_mpc.model.Model`)
         for each of these instances. The structure is initialized with all zeros.
         Use this object to define values of the time-varying parameters.
@@ -417,8 +571,8 @@ class Optimizer:
         :rtype: None
         """
 
-        tvp_template = struct_symSX([
-            entry('_tvp', repeat=self.n_horizon, struct=self.model._tvp)
+        tvp_template = self.model.sv.sym_struct([
+            entry('_tvp', repeat=self.n_horizon+1, struct=self.model._tvp)
         ])
         return tvp_template(0)
 
@@ -497,7 +651,7 @@ class Optimizer:
         assert self.flags['setup'] == True, 'optimizer was not setup yet. Please call optimizer.setup().'
 
         r = self.S(x0=self.opt_x_num, lbx=self.lb_opt_x, ubx=self.ub_opt_x,
-            ubg=self.cons_ub, lbg=self.cons_lb, p=self.opt_p_num)
+            ubg=self.nlp_cons_ub, lbg=self.nlp_cons_lb, p=self.opt_p_num)
         # Note: .master accesses the underlying vector of the structure.
         self.opt_x_num.master = r['x']
         self.opt_x_num_unscaled.master = r['x']*self.opt_x_scaling
@@ -540,12 +694,12 @@ class Optimizer:
         alg = substitute(alg, _z, _z*self._z_scaling.cat)
         alg = substitute(alg, _p, _p*self._p_scaling.cat) # only meaningful for MHE.
 
-        if self.state_discretization == 'discrete':
-            _i = SX.sym('i', 0)
+        if self.model.model_type == 'discrete':
+            _i = self.model.sv.sym('i', 0)
             # discrete integrator ifcs mimics the API the collocation ifcn.
             ifcn = Function('ifcn', [_x, _i, _u, _z, _tvp, _p, _w], [alg, rhs/self._x_scaling.cat])
             n_total_coll_points = 0
-        if self.state_discretization == 'collocation':
+        elif self.state_discretization == 'collocation':
             ffcn = Function('ffcn', [_x, _u, _z, _tvp, _p, _w], [rhs/self._x_scaling.cat])
             afcn = Function('afcn', [_x, _u, _z, _tvp, _p, _w], [alg])
             # Get collocation information
@@ -580,7 +734,7 @@ class Optimizer:
             D = np.zeros(deg + 1)
 
             # Dimensionless time inside one control interval
-            tau = SX.sym("tau")
+            tau = self.model.sv.sym("tau")
 
             # All collocation time points
             T = np.zeros((nk, ni, deg + 1))
@@ -606,24 +760,25 @@ class Optimizer:
                     C[j, r] = tfcn(tau_root[r])
 
             # Define symbolic variables for collocation
-            xk0 = SX.sym("xk0", n_x)
-            #zk = SX.sym("zk", n_z)
-            pk = SX.sym("pk", n_p)
-            tv_pk = SX.sym("tv_pk", n_tvp)
-            uk = SX.sym("uk", n_u)
-            wk = SX.sym("wk", n_w)
+            xk0 = self.model.sv.sym("xk0", n_x)
+            #zk = self.model.sv.sym("zk", n_z)
+            pk = self.model.sv.sym("pk", n_p)
+            tv_pk = self.model.sv.sym("tv_pk", n_tvp)
+            uk = self.model.sv.sym("uk", n_u)
+            wk = self.model.sv.sym("wk", n_w)
 
             # State trajectory
             n_ik = ni * (deg + 1) * n_x
-            ik = SX.sym("ik", n_ik)
-            ik_split = np.resize(np.array([], dtype=SX), (ni, deg + 1))
+            ik = self.model.sv.sym("ik", n_ik)
+
+            ik_split = np.resize(np.array([], dtype=self.model.sv.dtype), (ni, deg + 1))
             offset = 0
 
             # Algebraic trajectory
             n_zk = ni * (deg +1) * n_z
-            zk = SX.sym("zk", n_zk)
+            zk = self.model.sv.sym("zk", n_zk)
             offset_z = 0
-            zk_split = np.resize(np.array([], dtype=SX), (ni, deg + 1))
+            zk_split = np.resize(np.array([], dtype=self.model.sv.dtype), (ni, deg + 1))
 
             # Store initial condition
             ik_split[0, 0] = xk0
@@ -757,3 +912,52 @@ class Optimizer:
             'branch_offset': branch_offset
         }
         return n_branches, n_scenarios, child_scenario, parent_scenario, branch_offset
+
+
+    def prepare_nlp(self):
+        """Prepare the optimization problem.
+        Typically, this method is called internally from :py:meth:`setup`.
+
+        Users should only call this method if they intend to modify the objective with :py:attr:`nlp_obj`,
+        the constraints with :py:attr:`nlp_cons`, :py:attr:`nlp_cons_lb` and :py:attr:`nlp_cons_ub`.
+
+        To finish the setup process, users MUST call :py:meth:`create_nlp` afterwards.
+
+        .. note::
+
+            Do NOT call :py:meth:`setup` if you intend to go the manual route with :py:meth:`prepare_nlp` and :py:meth:`create_nlp`.
+
+        .. note::
+
+            Only AFTER calling :py:meth:`prepare_nlp` the previously mentionned attributes
+            :py:attr:`nlp_obj`, :py:attr:`nlp_cons`, :py:attr:`nlp_cons_lb`, :py:attr:`nlp_cons_ub`
+            become available.
+
+        """
+
+        # MPC and MHE have similar methods. The documentation is valid for both of them.
+        self._prepare_nlp()
+
+    def create_nlp(self):
+        """Create the optimization problem.
+        Typically, this method is called internally from :py:meth:`setup`.
+
+        Users should only call this method if they intend to modify the objective with :py:attr:`nlp_obj`,
+        the constraints with :py:attr:`nlp_cons`, :py:attr:`nlp_cons_lb` and :py:attr:`nlp_cons_ub`.
+
+        To finish the setup process, users MUST call :py:meth:`create_nlp` afterwards.
+
+        .. note::
+
+            Do NOT call :py:meth:`setup` if you intend to go the manual route with :py:meth:`prepare_nlp` and :py:meth:`create_nlp`.
+
+        .. note::
+
+            Only AFTER calling :py:meth:`prepare_nlp` the previously mentionned attributes
+            :py:attr:`nlp_obj`, :py:attr:`nlp_cons`, :py:attr:`nlp_cons_lb`, :py:attr:`nlp_cons_ub`
+            become available.
+
+        """
+
+        # MPC and MHE have similar methods. The documentation is valid for both of them.
+        self._create_nlp()
